from pyvista_gui._version import __version__
from pyvista_gui.console import *
from pyvista_gui.constants import *
from pyvista_gui.data import *
from pyvista_gui.gui import *
from pyvista_gui.options import *
from pyvista_gui.utilities import *
from pyvista_gui.widgets import *




import sys
from pydoc import help

# from qtconsole.manager import QtKernelManager
import qdarkstyle
from qtconsole.inprocess import QtInProcessKernelManager
from qtconsole.rich_jupyter_widget import RichJupyterWidget

FROZEN = getattr(sys, 'frozen', False)


def _complete(self):
    """ Performs completion at the current cursor location. """
    try:
        # Send the completion request to the kernel
        msg_id = self.kernel_client.complete(
            code=self.input_buffer,
            cursor_pos=self._get_input_buffer_cursor_pos(),
        )
        pos = self._get_cursor().position()
        info = self._CompletionRequest(msg_id, pos)
        self._request_info['complete'] = info
    except:
        pass


# disable completion when frozen
if FROZEN:
    from qtconsole import frontend_widget
    frontend_widget.FrontendWidget._complete = _complete


class QIPythonWidget(RichJupyterWidget):

    def __init__(self, parent, custom_banner=None, *args, **kwargs):
        super(QIPythonWidget, self).__init__(*args, **kwargs)
        if custom_banner is not None:
            self.banner = custom_banner

        self.font_size = 8
        self.default_style_sheet = self.styleSheet()

        self.kernel_manager = QtInProcessKernelManager()
        self.kernel_manager.start_kernel(show_banner=False)
        self.kernel_manager.kernel.gui = 'qt'
        self.kernel_client = self._kernel_manager.client()
        self.kernel_client.start_channels()
        self.gui = parent

        # override "exit" function
        def exit():
            """ Exit the pyvista_gui """
            parent.close()

        self.push_vars({'gui': parent,
                        'exit': exit,
                        'quit': exit,
                        'help': help})

    @property
    def variables(self):
        """ Variables local to the qtconsole """
        return self.shell.ns_table['user_local']

    def clear_variables(self):
        varlist = list(self.variables)
        for var in varlist:
            variable = self.variables[var]
            if hasattr(variable, 'remove'):
                try:
                    variable.remove()
                except:
                    pass

    @property
    def shell(self):
        """ Return shell object """
        return self.kernel_manager.kernel.shell

    def enable_dark_mode(self, state):
        if state:
            self.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
        else:
            self.setStyleSheet(self.default_style_sheet)

    def push_vars(self, variables):
        """Given a dictionary containing name / value pairs, push those
        variables to the Jupyter console widget.
        """
        self.shell.push(variables)

    def clear(self):
        """ Clear the terminal """
        self._control.clear()

    def execute_command(self, command):
        """Execute a command in the frame of the console widget"""
        self._execute(command, False)

    def _execute(self, source, hidden):
        """Additional layer on execution to save commands and stop the
        parent gui from doing the same.
        """
        # TODO:
        # if self.gui:
        #     self.gui.data.store_command(source)
        #     self.gui.save_commands = False

        super(QIPythonWidget, self)._execute(source, hidden)

        if self.gui:
            self.gui.save_commands = True



"""Widgets and associated tooltips"""
import logging

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import (QBoxLayout, QCheckBox, QColorDialog,
                             QDialog, QFileDialog, QFormLayout,
                             QFrame, QGroupBox, QLabel, QSlider,
                             QVBoxLayout)

LOG = logging.getLogger(__name__)
LOG.setLevel('DEBUG')


class ColorDialog(QColorDialog):
    """
    Color dialog query that emits a signal when a color is selected
    the dialog was property closed.
    """
    color_picked = pyqtSignal(tuple)

    def __init__(self, parent=None):
        super(ColorDialog, self).__init__(parent=parent)
        self.setOption(QColorDialog.DontUseNativeDialog)
        self.colorSelected.connect(self.trigger_accepted)

    def trigger_accepted(self, color):
        """
        Sends signal that the fem file dialog was closed properly.

        Sends:
        fem_filename, has_thickness, is_rotor
        """
        rgb = (color.red()/255.0, color.green()/255.0, color.blue()/255.0)
        self.color_picked.emit(rgb)


class FileDialog(QFileDialog):
    """Generic file query that emits a signal when a file is selected
    and the dialog was property closed.

    Examples
    --------

    >>> dlg = FileDialog(filefilter=["Text files (*.txt)", "Images (*.png *.jpg)"])
    """
    dlg_accepted = pyqtSignal(str)

    def __init__(self, parent=None, filefilter=None, save_mode=False,
                 show=True, callback=None, directory=False):
        super(FileDialog, self).__init__(parent)

        if filefilter is not None:
            self.setNameFilters(filefilter)
        self.setOption(QFileDialog.DontUseNativeDialog)
        self.accepted.connect(self.emit_accepted)
        if show:
            self.show()

        if directory:
            self.FileMode(QFileDialog.DirectoryOnly)
            self.setOption(QFileDialog.ShowDirsOnly, True)

        if save_mode:
            self.setAcceptMode(QFileDialog.AcceptSave)

        if callback is not None:
            self.dlg_accepted.connect(callback)

    def emit_accepted(self):
        """
        Sends signal that the file dialog was closed properly.

        Sends:
        filename
        """
        if self.result():
            self.dlg_accepted.emit(self.selectedFiles()[0])


class ColorBox(QFrame):

    colorChanged = pyqtSignal(tuple)

    def __init__(self, parent=None, rgb=[1, 1, 1]):
        super(ColorBox, self).__init__(parent)

        self.bgColor = Qt.white
        self.setStyleSheet("margin:5px; border:1px; ")

        self.set_background(self.upscale_rgb(rgb))

    def mousePressEvent(self, event):
        if event.buttons() == Qt.LeftButton:
            col = QColorDialog.getColor(self.bgColor, self)
            if col.isValid():
                rgb = (col.red(), col.green(), col.blue())
                self.set_background(rgb)
                self.bgColor = col
                self.colorChanged.emit(self.downscale_rgb(rgb))

    def upscale_rgb(self, rgb):
        return (int(rgb[0]*255), int(rgb[1]*255), int(rgb[2]*255))

    def downscale_rgb(self, rgb):
        return (rgb[0]/255.0, rgb[1]/255.0, rgb[2]/255.0)

    def set_background(self, rgb):
        self.setStyleSheet("QWidget { background-color: rgb(%d,%d,%d); border:1px solid rgb(0, 0, 0); }" % rgb)


class LoadMeshDialog(QDialog):
    """ Custom file dialog to open up PyVista-supported files """
    accepted = pyqtSignal(str, bool)

    def __init__(self, parent=None, show=True, callback=None):
        super().__init__(parent)
        main_dialog = QDialog()
        main_dialog.setWindowTitle('Select CFD File')
        layout = QVBoxLayout(main_dialog)

        self.file_dialog = QFileDialog(main_dialog, Qt.Widget)
        # self.file_dialog.setNameFilters(FILE_FILTER)
        self.file_dialog.setWindowFlags(self.file_dialog.windowFlags() & ~Qt.Dialog)
        self.file_dialog.setOption(QFileDialog.DontUseNativeDialog)

        # close all when main closes
        self.file_dialog.finished.connect(self.close)
        layout.addWidget(self.file_dialog)

        form_layout = QFormLayout(self)
        settings_groupbox = QGroupBox('Settings')

        self.isrotor_checkbox = QCheckBox(self)
        self.isrotor_checkbox.setChecked(False)
        self.isrotor_checkbox.setToolTip('CFD grids are a sector of a rotor')
        form_layout.addRow('Sector', self.isrotor_checkbox)

        settings_groupbox.setLayout(form_layout)
        layout.addWidget(settings_groupbox)
        self.setLayout(layout)
        self.file_dialog.finished.connect(self.emit_accepted)

        if show:
            self.show()

        if callback is not None:
            self.accepted.connect(callback)

    @property
    def is_rotor(self):
        return self.isrotor_checkbox.isChecked()

    def emit_accepted(self, result):
        """Sends signal that the fem file dialog was closed properly.

        Sends:
        cfd_filename, is_rotor
        """
        if result:
            self.accepted.emit(self.file_dialog.selectedFiles()[0],
                               self.is_rotor)


class SlidersGroup(QGroupBox):

    nocheckvalueChanged = pyqtSignal(int)
    fixedvalueChanged = pyqtSignal(int)

    def __init__(self, orientation, title, parent=None):
        super(SlidersGroup, self).__init__(title, parent)

        def MakeSlider(orientation):
            slider = QSlider(orientation)
            slider.setFocusPolicy(Qt.StrongFocus)
            slider.setTickPosition(QSlider.TicksBothSides)
            slider.setTickInterval(100)
            slider.setSingleStep(1)
            return slider

        # nocheck slider
        self.nocheck_slider = MakeSlider(orientation)
        self.nocheck_slider.valueChanged.connect(self.nocheckvalueChanged)

        self.fixed_slider = MakeSlider(orientation)
        self.fixed_slider.valueChanged.connect(self.fixedvalueChanged)

        slidersLayout = QBoxLayout(QBoxLayout.TopToBottom)
        slidersLayout.addWidget(self.nocheck_slider)
        slidersLayout.addWidget(self.fixed_slider)
        slidersLayout.addWidget(QLabel(''))
        self.setLayout(slidersLayout)

    def setEnabled(self, value):
        self.fixed_slider.setEnabled(value)
        self.nocheck_slider.setEnabled(value)

    def set_fixed_value(self, value):
        value = value*1000
        self.fixed_slider.setValue(value)
        if self.nocheck_slider.value() > value:
            self.nocheck_slider.setValue(value)

    def set_nocheck_value(self, value):
        value = value*1000
        self.nocheck_slider.setValue(value)
        if self.fixed_slider.value() < value:
            self.fixed_slider.setValue(value)

    def setMinimum(self, value):
        self.fixed_slider.setMinimum(value*1000)
        self.nocheck_slider.setMinimum(value*1000)

    def setMaximum(self, value):
        self.fixed_slider.setMaximum(value*1000)
        self.nocheck_slider.setMaximum(value*1000)



"""Options for saving user prefences, etc.
"""
import json
import os

import pyvista


class RcParams(dict):
    """Internally used class to manage the rcParams"""

    filename = os.path.join(pyvista.USER_DATA_PATH, 'rcParams.json')

    def save(self):
        with open(self.filename, 'w') as f:
            json.dump(self, f)
        return

    def load(self):
        with open(self.filename, 'r') as f:
            data = json.load(f)
        self.update(data)

    def __setitem__(self, key, value):
        dict.__setitem__(self, key, value)
        self.save()

# The options
rcParams = RcParams(
    dark_mode=False,
)

# Load user prefences from last session if none exist, save defaults
try:
    rcParams.load()
except:
    rcParams.save()



""" This is the main GUI """

import datetime
import logging

import qdarkstyle
from PyQt5.QtCore import Qt, pyqtSignal
# from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (QAction, QDockWidget, QHBoxLayout,
                             QMainWindow, QMenu, QMessageBox, QProgressDialog,
                             QSizePolicy, QSplitter, QVBoxLayout)

from pyvista.plotting import Plotter
from pyvistaqt import QtInteractor

from pyvista_gui.console import QIPythonWidget
from pyvista_gui.data import Data
from pyvista_gui.dialogs import ColorDialog, LoadMeshDialog
from pyvista_gui.options import rcParams
from pyvista_gui.widgets import QTextEditCommands, QTextEditLogger, TreeWidget

# from weakref import proxy
# from functools import wraps


logging.getLogger('ipykernel').setLevel('CRITICAL')
logging.getLogger('traitlets').setLevel('CRITICAL')
logging.getLogger('root').setLevel('CRITICAL')
logging.getLogger('parso').setLevel('CRITICAL')
logging.getLogger('parso.cache').setLevel('CRITICAL')

LOG = logging.getLogger(__name__)
LOG.setLevel('DEBUG')


NOW = datetime.datetime.now()
ABOUT_TEXT = """PyVista Graphical User Interface
Copyright %d

Written by the PyVista Developers
info@pyvista.org

""" % NOW.year


CHANGE_BACKGROUND_MENU_TEXT = 'Change Background'


###############################################################################

class GUIWindow(QMainWindow):
    """ Primary GUI object """
    trigger_render = pyqtSignal()
    errorsignal = pyqtSignal(str, str)
    closepbar_signal = pyqtSignal()

    def __init__(self, parent=None, app=None, show=True, off_screen_vtk=False):
        """ Generate window and initialize a scene """
        QMainWindow.__init__(self, parent)
        LOG.debug('Initializing GUI...')
        self.app = app
        self.default_style_sheet = self.styleSheet()
        self.save_commands = True  # stores commands internally when True
        self.hold = False
        self.off_screen_vtk = off_screen_vtk
        self.load_dialog = None


        self.resize(800, 600)
        self.setWindowTitle('PyVista GUI')

        # tree and associated queries
        self.data = Data(self)

        # Create docks

        # ipython console
        self.console = QIPythonWidget(self)
        self.dock_console = QDockWidget('Console', self)
        self.dock_console.setWidget(self.console)
        self.addDockWidget(Qt.BottomDockWidgetArea, self.dock_console)

        # commands
        self.textedit_commands = QTextEditCommands(self)
        self.dock_commands = QDockWidget('Commands', self)
        self.dock_commands.setWidget(self.textedit_commands)
        # self.addDockWidget(Qt.BottomDockWidgetArea, self.dock_commands)

        # object tree
        self.tree = TreeWidget(self)
        self.dock_tree = QDockWidget('Object Tree', self)
        self.dock_tree.setWidget(self.tree)
        self.addDockWidget(Qt.LeftDockWidgetArea, self.dock_tree)

        # log panel
        self.textbox_logger = QTextEditLogger(self)
        formatter = logging.Formatter('%(name)-20s - %(levelname)-8s - %(message)s')
        self.textbox_logger.setFormatter(formatter)
        logging.getLogger().addHandler(self.textbox_logger)
        logging.getLogger().setLevel('DEBUG')
        self.dock_logger = QDockWidget('Log', self)
        self.dock_logger.setWidget(self.textbox_logger.widget)
        # self.addDockWidget(Qt.BottomDockWidgetArea, self.dock_logger)

        # vtk frame if available
        if off_screen_vtk:
            self.plotter = Plotter(off_screen=True)
        else:
            self.plotter = QtInteractor(self)
            self.dock_vtk = QDockWidget('Viewer', self)
            self.dock_vtk.setWidget(self.plotter.interactor)
            self.addDockWidget(Qt.RightDockWidgetArea, self.dock_vtk)

            self.splitDockWidget(self.dock_tree, self.dock_vtk, Qt.Horizontal)
            self.splitDockWidget(self.dock_vtk, self.dock_console, Qt.Vertical)

            self.resizeDocks([self.dock_vtk, self.dock_console], [4, 1], Qt.Vertical)
            self.resizeDocks([self.dock_vtk, self.dock_tree], [4, 1],
                             Qt.Horizontal)

            # self.plotter.add_toolbars(self)

        self.tabifyDockWidget(self.dock_console, self.dock_commands)
        self.tabifyDockWidget(self.dock_commands, self.dock_logger)
        self.dock_console.raise_()

        # Create menu
        self.make_menu()

        # connects
        self.trigger_render.connect(self.plotter.render)
        self.errorsignal.connect(self.error_dialog)
        self.closepbar_signal.connect(self._closepbar)
        LOG.debug('GUI initialized')

        # Show frame
        self.enable_dark_mode(rcParams['dark_mode'])

        if show:
            self.show()


    def make_menu(self):
        """ Generates menus """
        self.menu = self.menuBar()
        self.menu.setNativeMenuBar(False)

        # main menu
        # self.file_menu = self._build_file_menu()
        self.view_menu = self._build_view_menu()
        # self.panels_menu = self._build_panels_menu()
        # self.help_menu = self._build_help_menu()


    def error_dialog(self, txt, textinfo=None):
        """ Creates error dialogue """
        self.err_message_box = QMessageBox(self)
        self.err_message_box.setIcon(QMessageBox.Critical)

        self.err_message_box.setText(txt)
        if textinfo:
            self.err_message_box.setInformativeText(textinfo)
        self.err_message_box.setWindowTitle('Error')
        self.err_message_box.show()

    def _closepbar(self):
        """ Only to be accessed by a signal call """
        self.pbar.signal_close()

    def _build_view_menu(self):
        """ Creates view menu """
        menu = self.menu.addMenu('View')
        if not self.off_screen_vtk:
            self.add_menu_item(menu, CHANGE_BACKGROUND_MENU_TEXT,
                               self.change_background)

            # Camera views
            sub_menu = QMenu('Camera Positions', parent=self)
            menu.addMenu(sub_menu)

            _view_vector = lambda *args: self.plotter.view_vector(*args)
            cvec_setters = {
                # Viewing vector then view up vector - PyVista handles the rest
                'Top (-Z)': lambda: _view_vector((0,0,1), (0,1,0)),
                'Bottom (+Z)': lambda: _view_vector((0,0,-1), (0,1,0)),
                'Front (-Y)': lambda: _view_vector((0,1,0), (0,0,1)),
                'Back (+Y)': lambda: _view_vector((0,-1,0), (0,0,1)),
                'Left (-X)': lambda: _view_vector((1,0,0), (0,0,1)),
                'Right (+X)': lambda: _view_vector((-1,0,0), (0,0,1)),
            }

            for key in cvec_setters.keys():
                self.add_menu_item(sub_menu, key, cvec_setters[key])

            action = QAction('Anti-Aliasing', menu, checkable=True)
            action.setChecked(True)
            action.triggered.connect(self.plotter.enable_anti_aliasing)
            menu.addAction(action)

        self.action_dark_mode = QAction('Dark Mode', menu, checkable=True)
        self.action_dark_mode.setChecked(rcParams['dark_mode'])
        self.enable_dark_mode(rcParams['dark_mode'])
        self.action_dark_mode.triggered.connect(self.enable_dark_mode)
        menu.addAction(self.action_dark_mode)

        self.view_menu = menu


    def enable_dark_mode(self, state=True):
        """ sets style sheet to dark mode """
        if state:
            self.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
        else:
            self.setStyleSheet(self.default_style_sheet)
        self.console.enable_dark_mode(state)
        rcParams['dark_mode'] = state

        if hasattr(self, 'action_dark_mode'):
            self.action_dark_mode.blockSignals(True)
            self.action_dark_mode.setChecked(state)
            self.action_dark_mode.blockSignals(False)


    def change_background(self):
        """ Pulls up change background dialog """
        self.color_dlg = ColorDialog(self)
        self.color_dlg.color_picked.connect(self.plotter.set_background)


    def add_menu_item(self, menu, text, func, addsep=False, enabled=True,
                      shortcut=None):
        """ Clean way of adding menu items """
        item = QAction(text, self)
        if func:
            item.triggered.connect(func)

        if addsep:
            menu.insertSeparator(menu.addAction(item))
        else:
            menu.addAction(item)

        if enabled:
            item.setEnabled(True)
        else:
            item.setEnabled(False)

        if shortcut is not None:
            item.setShortcut(shortcut)

        return item


    def load_mesh(self):
        """Loads a mesh from file using a file dialog"""
        self.file_dialog = LoadMeshDialog(self, callback=self.data.load_meesh)



import logging
import sys
import time
import traceback
from functools import wraps
from threading import Thread

log = logging.getLogger(__name__)
log.setLevel('DEBUG')


# python 2/3 string compatability
if sys.version[0] == '2':
    from past.builtins import basestring
else:
    basestring = str


def build_command(obj, func, *args, **kwargs):
    """ build a command for a non-gui session of pyvista """
    import pyvista

    func_name = func.__name__
    if func_name[0] == '_':
       func_name = func_name[1:]

    # verify function is a method in a morph class
    if not hasattr(obj, 'class_name'):
        return

    if hasattr(pyvista, obj.class_name):
        pyvista_class = eval('pyvista.%s' % obj.class_name)
    else:
        return

    if not hasattr(pyvista_class, func_name):
        return

    # convert args to strings
    varname = obj.varname
    str_args = []
    for arg in args[1:]:
        # if isinstance(arg, GuiCommon):
        #     str_args.append(arg.varname)
        if isinstance(arg, basestring):
            str_args.append('"%s"' % arg)
        else:
            str_args.append(str(arg))

    return '%s.%s(%s)' % (varname, func_name, ', '.join(str_args))


def protected_thread(fn):
    """
    Calls a function using a thread.  Reports error under the
    assumption the first argument is a GuiCommon class.
    """
    @wraps(fn)
    def wrapper(*args, **kwargs):
        self = args[0]

        def protected_fn():
            try:
                fn(*args, **kwargs)
                command = build_command(self, fn, *args, **kwargs)
                self.store_command(command)
            except Exception as exception:
                exc_info = sys.exc_info()
                traceback.print_exception(*exc_info)
                log.error(exception)
                if hasattr(self, 'exceptions'):
                    self.exceptions.append(exception)
                self.parent.show_error(exception)

        thread = Thread(target=protected_fn)
        thread.start()

        if self.parent.hold:
            while thread.is_alive():
                # self.parent.render()
                self.parent.app.processEvents()
                time.sleep(0.1)

        if hasattr(self, 'threads'):
            self.threads.append(thread)
        return thread

    # wrapper.__doc__ = fn.__doc__
    # wrapper.__name__ = fn.__name__

    return wrapper


def threaded(fn):
    """ calls a function using a thread """
    def wrapper(*args, **kwargs):
        thread = Thread(target=fn, args=args, kwargs=kwargs)
        thread.start()
        return thread
    return wrapper



import logging
import os
import sys

from PyQt5.QtWidgets import QApplication

from pyvista_gui.gui import GUIWindow

LOG = logging.getLogger(__name__)
LOG.setLevel('DEBUG')



def main(debug=False, loglevel='DEBUG', script=None, off_screen_vtk=False):  # pragma: no cover
    """ Starts the PyVista GUI """

    logging.getLogger().setLevel('CRITICAL')

    app = QApplication(sys.argv)
    gui = GUIWindow(app=app, off_screen_vtk=off_screen_vtk)

    # icon_file = resource_path('icon.ico')
    # if os.path.isfile(icon_file):
    #     icon = QIcon(icon_file)
    #     gui.setWindowIcon(icon)
    # else:
    #     LOG.warning('Unable to find icon file')

    # if script is not None:
    #     if os.path.isfile(script):
    #         exec('gui.data.load_script("%s")' % script)
    #     else:
    #         gui.show_error('Unable to find python script file "%s"\n' % str(script)
    #                        + 'Try using absolute paths or loading it from '
    #                        + 'the GUI.')

    # always start in home directory
    try:
        from pathlib import Path
        home = str(Path.home())
        os.chdir(home)
    except:
        LOG.warning('Unable to change to home directory')

    app.exec_()





if __name__ == '__main__':
    main()



"""
Installation file for python pyvista_gui module
"""
import os
import platform
import sys
import warnings
from io import open as io_open

from setuptools import setup

package_name = 'pyvista_gui'

__version__ = None
filepath = os.path.dirname(__file__)
version_file = os.path.join(filepath, package_name, '_version.py')
with io_open(version_file, mode='r') as fd:
    exec(fd.read())

# pre-compiled vtk available for python3
install_requires = [
    'pyvista>=0.22.4',
    'PyQt5>=5.11.3',
    'matplotlib',
    'pyvistaqt',
    'qdarkstyle',
    'qtconsole'
]

readme_file = os.path.join(filepath, 'README.md')

setup(
    name=package_name,
    packages=[package_name],
    version=__version__,
    description='Easier Pythonic interface to VTK',
    long_description=io_open(readme_file, encoding="utf-8").read(),
    long_description_content_type="text/x-rst",
    author='PyVista Developers',
    author_email='info@pyvista.org',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    url='https://github.com/pyvista/pyvista_gui',
    keywords='vtk numpy plotting mesh gui',
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
    install_requires=install_requires,
)



import logging

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtGui import QFont, QStandardItem, QStandardItemModel
from PyQt5.QtWidgets import QPlainTextEdit, QTreeView


class QTextEditLogger(QPlainTextEdit, logging.Handler):
    write_text_signal = pyqtSignal(str)

    def __init__(self, parent):
        super().__init__()
        self.widget = QPlainTextEdit(parent)
        self.widget.setReadOnly(True)
        self.write_text_signal.connect(self.widget.appendPlainText)

        self.widget.setFont(QFont("Courier", 12))

    def emit(self, record):
        msg = self.format(record)
        self.write_text_signal.emit(msg)


class QTextEditCommands(QPlainTextEdit):
    """ Displays commands as they are run by the GUI """
    write_text_signal = pyqtSignal(str)

    def __init__(self, parent):
        super(QTextEditCommands, self).__init__(parent)
        self.setReadOnly(True)
        self.write_text_signal.connect(self.appendPlainText)

        self.setFont(QFont("Courier", 12))

    def add_command(self, command):
        """ Writes a command to the panel """
        if command:
            self.write_text_signal.emit('%s\n' % command)


class TreeWidget(QTreeView):
    """ Displays objects as an object tree """

    def __init__(self, parent):
        super(TreeWidget, self).__init__(parent)
        self.parent = parent

        # add menu
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.open_menu)

        self.model = QStandardItemModel()
        self.setModel(self.model)
        self.model.setHorizontalHeaderLabels(["Objects"])
        self.model.itemChanged.connect(self.edit_name)

    def edit_name(self, item):
        """ Makes the name of the item in the tree matches the object name """
        obj = item.data()
        if obj:
            obj.name = item.text()

    def addItem(self, item, header, mainheader=None, editable=True):
        """ Adds item to tree """
        if mainheader is None:
            finds = self.model.findItems(header)
            if not finds:
                itemheader = QStandardItem(header)
                self.model.appendRow(itemheader)
            else:
                itemheader = finds[0]
        else:
            finds = self.model.findItems(mainheader)
            if not finds:
                raise Exception('main header %s does not exist' % mainheader)
            main_itemheader = finds[0]

            for i in range(main_itemheader.rowCount()):
                itemheader = main_itemheader.child(i, 0)
                if itemheader:
                    if header is itemheader.data():
                        break

        standardItem = QStandardItem(item.name)
        standardItem.setEditable(editable)
        standardItem.setData(item)
        itemheader.appendRow(standardItem)

    def open_menu(self, position):  # pragma: no cover
        """ Activates when right click in tree """
        index = self.selectedIndexes()

        if not index:
            return
        item = self.model.itemFromIndex(index[0])
        obj = item.data()

        if hasattr(obj, 'menu'):
            menu = obj.menu
            menu.exec_(self.viewport().mapToGlobal(position))
            return menu

    def remove_item(self, data):
        """ removes an item from the modal list """
        removed = False
        # check each header for a match
        for i in range(self.model.rowCount()):
            header = self.model.item(i)
            for i in range(header.rowCount()):
                item = header.child(i, 0)
                if item:
                    if data is item.data():
                        item.setData(None)
                        header.removeRow(i)
                        removed = removed or True
                        break

                if item.hasChildren():
                    for j in range(item.rowCount()):
                        subitem = item.child(j, 0)
                        if data is subitem.data():
                            subitem.setData(None)
                            item.removeRow(j)
                            removed = removed or True
                            break

            # remove empty headers
            if not header.rowCount():
                self.model.removeRow(header.row())
                break

        return removed



"""pyvista_gui data module"""
import logging
import os

# from PyQt5.QtWidgets import QInputDialog
from PyQt5.QtWidgets import (QApplication, QBoxLayout, QCheckBox, QColorDialog,
                             QComboBox, QDialog, QDoubleSpinBox, QFileDialog,
                             QFormLayout, QFrame, QGridLayout, QGroupBox,
                             QHBoxLayout, QLabel, QListWidget, QMenu,
                             QPushButton, QSlider, QSpinBox, QStackedWidget,
                             QVBoxLayout)

import pyvista

import pyvista_gui
from pyvista_gui.constants import PY_FILE_FILTER
from pyvista_gui.dialogs import FileDialog
from pyvista_gui.utilities import protected_thread

LOG = logging.getLogger(__name__)
LOG.setLevel('DEBUG')


class Data():
    """ Holds active data for the gui """

    def __init__(self, parent):
        self.parent = parent
        self.meshes = []
        self.commands = []
        self.load_script_dlg = None

        # initialize cached commands
        self.reset_stored_commands()

        self.varcount = {'Mesh': 0,}

    def reset_stored_commands(self):
        """resets stored commands"""
        self.commands = []

    def list_commands(self):
        for command in self.commands:
            print(command)

    @property
    def items(self):
        items = []
        items.extend(self.meshes)
        return items

    def save_commands_dialog(self):
        """ Saves commands to a python script """
        self.save_commands_dlg = FileDialog(self.parent,
                                            PY_FILE_FILTER,
                                            save_mode=True,
                                            callback=self._save_commands)
        return self.save_commands_dlg

    def _save_commands(self, filename):
        """ Save commands to file """
        header = '"""\nAuto-generated script for PyVista-GUI\n'
        header += 'Using pyvista     v%s\n' % pyvista.__version__
        header += '      pyvista_gui v%s\n' % pyvista_gui.__version__
        header += '"""\n'

        # determine extra imports
        import_modules = []
        test_modules = []
        for command in self.commands:
            for module in test_modules:
                if '%s.' % module in command:
                    if module not in import_modules:
                        import_modules.append(module)

        for module in import_modules:
            self.commands.insert(0, 'import %s' % module)

        LOG.info('Writing auto-generated script for PyVista to %s', filename)

        with open(filename, 'w') as f:
            f.write(header + '\n')
            for command in self.commands:
                if command is not None:
                    command = command.replace('\\', '/')
                    f.write(command + '\n')

    def store_command(self, command):
        """ Stores a command """
        if command is not None:
            self.commands.append(command)

    @protected_thread
    def load_mesh(self, uinput, name=None, reset_camera=True):
        """ Adds a surface to self """
        raise NotImplementedError()
        # TODO
        # return GuiSurface(uinput, self.parent, name=name, reset_camera=reset_camera)



    def remove(self, item):
        """ Removes an item from the database """
        if item in self.meshes:
            self.meshes.remove(item)


    def reset(self):
        """ removes all items from database """
        for item in self.items:
            item.remove()

        # reset commands
        self.reset_stored_commands()
        if hasattr(self.parent.plotter, 'reset'):
            self.parent.plotter.reset()
        self.parent.textedit_commands.clear()

        # clear gui widgets
        self.parent.textbox_logger.widget.clear()
        self.parent.console.clear()

    # def save(self, filename):  # pragma: no cover
    #     """ save all data """
    #     with zipfile.ZipFile(filename, 'w') as zipfile:
    #         for item in self.items:
    #             item.write_to_zip(zipfile)

    def load_script_dialog(self):
        """ Open up a file dialog to load a Python script """
        self.load_script_dlg = FileDialog(self.parent, PY_FILE_FILTER,
                                          callback=self.load_script)
        return self.load_script_dlg

    def load_script(self, filename):
        """ Run a Python script from the qtconsole """
        if not os.path.isfile(filename):
            err_str = 'Unable to find python script file "%s"\n' % str(filename) +\
                      'Try using absolute paths or loading it from '+\
                      'the GUI.'
            LOG.error(err_str)
            self.parent.show_error(err_str)
            return

        # execute script and don't save internal commands
        self.parent.save_commands = False
        self.parent.hold = True

        # set hold parameter to False when complete to let main know
        # the script is complete
        command = 'exec(open("%s").read()); gui.hold = False' % filename
        command += ';gui.save_commands = True'
        self.parent.console.execute(command)

        with open(filename) as f:
            text = f.read()
        self.store_command(text)

        note = '# commands from %s' % filename
        self.parent.textedit_commands.add_command(note)
        self.parent.textedit_commands.add_command(text)
        note = '# finished with commands from %s' % filename
        self.parent.textedit_commands.add_command(note)
